/*
 * MultipleAlignment.cpp
 *
 *  Created on: Aug 12, 2009
 *      Author: kohany
 */

#include <stdexcept>
#include <iomanip>
#include "MultipleAlignment.h"

//#include "MSS.h"
//using namespace mss;

#ifndef _NO_ALN
#include "ThreadedAligner.h"
ThreadedAligner MultipleAlignment::aligner;
#endif

using namespace std;

bool MultipleAlignment::DEBUG = false;

MultipleAlignment::MultipleAlignment() : score(0)
{
   // TODO Auto-generated constructor stub
}
MultipleAlignment::MultipleAlignment(const MultipleAlignment& other) :
   alnvec(other.begin(), other.end()), score(other.score)
{
   // TODO Auto-generated destructor stub
}
MultipleAlignment::~MultipleAlignment()
{
   // TODO Auto-generated destructor stub
}
MultipleAlignment& MultipleAlignment::operator = (const MultipleAlignment& other) {
   alnvec.assign(other.begin(), other.end());
   score = other.getScore();
}

#ifndef _NO_ALN
void MultipleAlignment::setGapIni(int gapIni) {
   aligner.setGapIni(gapIni);
}
void MultipleAlignment::setGapExt(int gapExt) {
   aligner.setGapExt(gapExt);
}
void MultipleAlignment::setAlignmentMatrix(const ScoreMatrix& mat) {
   aligner.setAlignmentMatrix(mat);
}
void MultipleAlignment::setAlignmentMethod(ALNALG method) {
   aligner.setAlignmentMethod(method);
}
#endif

void MultipleAlignment::clear() {
   alnvec.clear();
   score = 0;
}
void MultipleAlignment::autoConsensus(BioSequence& cons, const ScoreMatrix* conmatptr, int minCnt, TAXON species, bool finalize, bool keep_gaps) const
{
   getConsensus(cons, conmatptr, minCnt);
   cons.setName(front().getName());
   if (finalize)
   {
      /**
       * Reconstructing CpG doublets...
       */
      if (species != UNKNOWN)
      {
         restoreCpG(cons, cons, species);
      }
#ifndef _NO_ALN
      //correctLowScoringRegions(sCON, conmatptr);
#endif
   }
   if(!keep_gaps) {
      cons.removeGaps();
   }
}

void MultipleAlignment::getConsensus(string& cons, const ScoreMatrix* matptr, int minNonGapCount) const
{
   getConsensus(cons, this->begin(), this->end(), matptr, minNonGapCount);
}

void MultipleAlignment::subAlignment(MultipleAlignment& subaln, size_t pos, size_t length) const {
   for(MultipleAlignment::const_iterator ptr = begin(); ptr != end(); ptr++) {
      subaln.push_back((*ptr).subseq(pos, length, true));
   }
}
/** ======================================================== */
double MultipleAlignment::calcAlignmentScores(const string& ref, vector<double>& columnScores, vector<double>& rowScores, const ScoreMatrix* matrix) const {
   double sumScore = 0;
   double score = 0;
   if (size() == 0)
   {
      return 0;
   }
   columnScores.assign(front().size(), 0);
   rowScores.assign(size(), 0);
   for (int i = 0; i < size(); i++)
   {
      score = BioAlignment::calcScore(ref, alnvec[i], 0, ref.size(), matrix, &columnScores, true, false);
      rowScores[i] += score;
      sumScore += score;
   }
   return sumScore;
}
/** ======================================================== */
double MultipleAlignment::calcAlignmentScores(vector<double>& columnScores, vector<double>& rowScores, const ScoreMatrix* matrix) const {
   double sumScore = 0;
   double score = 0;
   if (size() == 0)
   {
      return 0;
   }
   columnScores.assign(front().size(), 0);
   rowScores.assign(size(), 0);
   for (int i = 0; i < size(); i++)
   {
      int j = i;
      for (j++; j < size(); j++)
      {
         score = BioAlignment::calcScore(alnvec[i], alnvec[j], 0, alnvec[i].size(), matrix, &columnScores, true, false);
         rowScores[j] += score;
         rowScores[i] += score;
         sumScore += score;
      }
   }
   return sumScore;
}
/** ======================================================== */
/*
#ifndef _NO_ALN
void MultipleAlignment::correctLowScoringRegions(BioSubSequence& sCON, const ScoreMatrix* matptr) const
{
   typedef vector<double> SCOREVEC;
   typedef SCOREVEC::iterator SCOREPTR;
   typedef pair<vector<double>::iterator, vector<double>::iterator> SCOREPTRRANGE;
   list<SCOREPTRRANGE> badRegions;
   SCOREVEC columnScores(size(), 0);
   SCOREVEC rowScores(front().size(), 0);
   if(sCON.size() != front().size()) {
      throw invalid_argument("Specified consensus is not of the same length as alignment.");
   }

   double sumScore = calcAlignmentScores(sCON, columnScores, rowScores, matptr);
   sumScore = 0;
   for (SCOREPTR scoreptr = columnScores.begin(); scoreptr != columnScores.end(); scoreptr++)
   {
      *scoreptr = -(*scoreptr);
      sumScore += *scoreptr;
   }
   //double scoreThreshold = sumScore / columnScores.size();
   double scoreThreshold = 0;
   //cerr << "Autothreshold=" << scoreThreshold << endl;
   AlgMSS(columnScores.begin(), columnScores.end(), back_insert_iterator<list<SCOREPTRRANGE> >(badRegions), scoreThreshold);
   for (list<SCOREPTRRANGE>::reverse_iterator rgnptr = badRegions.rbegin(); rgnptr != badRegions.rend(); rgnptr++)
   {
      size_t pos = distance(columnScores.begin(), (*rgnptr).first);
      size_t len = (*rgnptr).second - (*rgnptr).first;
      //cerr << "Low scoring region: " << pos << ":" << len << " " << sCON.substr(pos, len) << endl;
      MultipleAlignment oldMaln;
      subAlignment(oldMaln, pos, len);
      for(MultipleAlignment::iterator ptr = oldMaln.begin(); ptr != oldMaln.end(); ptr++) {
         (*ptr).removeGaps();
      }

      BioSubSequence ref;
      MultipleAlignment newMaln;
      MultipleAlignment::iterator refptr = newMaln.autoAlign(oldMaln.begin(), oldMaln.end());
      if(refptr != oldMaln.end()) {
         newMaln.autoConsensus(ref, matptr, 2, UNKNOWN, true);
         if (sCON.substr(pos, len) != ref)
         {
            cerr << "Replacing " << pos << ":" << len << " : " << sCON.substr(pos, len) << " with " << ref.c_str() << endl;
            sCON.replace(pos, len, ref.c_str());
         }
      }
      else {
         cerr << "Erasing " << pos << ":" << len << " : " << sCON.substr(pos, len) << endl;
         sCON.erase(pos, len);
         //cerr << "Failed to rebuild consensus for positions " << pos << ":" << len << endl;
      }
   }
}
#endif
*/

double MultipleAlignment::getScore() const {
   return score;
}

int MultipleAlignment::read(istream& is, bool trimGaps) {
   int cnt = 0;
   BioSubSequence seq;
   while(seq.read(is) >= 0) {
      if (trimGaps)
      {
         for (BioSubSequence::iterator sptr = seq.begin(); sptr != seq.end(); sptr++)
         {
            if (*sptr == ScoreMatrix::GAPCHAR)
            {
               *sptr = ScoreMatrix::LPADDING;
            }
            else
            {
               break;
            }
         }
         for (BioSubSequence::reverse_iterator sptr = seq.rbegin(); sptr != seq.rend(); sptr++)
         {
            if (*sptr == ScoreMatrix::GAPCHAR)
            {
               *sptr = ScoreMatrix::RPADDING;
            }
            else
            {
               break;
            }
         }
      }
      push_back(seq);
      cnt++;
   }
   return cnt;
}

int MultipleAlignment::print(ostream& os) {
   for(MultipleAlignment::iterator ptr = begin(); ptr != end(); ptr++) {
      (*ptr).print(os, false);
   }
}

void MultipleAlignment::restoreCpG(const BioSequence& icons, BioSequence& ocons, TAXON species) const {
   restoreCpG(icons, ocons, begin(), end(), species);
}

/**
 * Probability of transition-mutation based on average similarity to
 * consensus. This implementation is biologically correct for mammals only.
 */
double inline pts_mam(double sim)
{
   return 2.0 / 3.0 * (1.0 - sim);
}
double inline pCA_TG(double pTs, double Sc)
{
   return pTs * pTs + pTs * Sc / 2;
}
double inline pTG_CA(double pTs, double Sc)
{
   return pCA_TG(pTs, Sc);
}
double inline pCA_TA(double pTs, double Sc)
{
   return pTs * Sc + pTs * pTs / 2;
}
double inline pTG_TA(double pTs, double Sc)
{
   return pCA_TA(pTs, Sc);
}
double inline pCA_CA(double pTs, double Sc)
{
   return Sc * (Sc + pTs / 2);
}
double inline pTG_TG(double pTs, double Sc)
{
   return pCA_CA(pTs, Sc);
}
double inline pTG_CG(double pTs, double Sc)
{
   return Sc * Sc / 2;
}
double inline pCA_CG(double pTs, double Sc)
{
   return pTG_CG(pTs, Sc);
}
double inline pTA_YR(double pTs, double Sc)
{
   return Sc * Sc / 4 + pTs * Sc;
}
double inline pTG_CR(double pTs, double Sc)
{
   return (pTG_CA(pTs, Sc) + pTG_CG(pTs, Sc)) / 2;
}
double inline pTG_TR(double pTs, double Sc)
{
   return (pTG_TA(pTs, Sc) + pTG_TG(pTs, Sc)) / 2;
}
double inline pCA_TR(double pTs, double Sc)
{
   return (pCA_TA(pTs, Sc) + pCA_TG(pTs, Sc)) / 2;
}
double inline pCA_YA(double pTs, double Sc)
{
   return (pCA_TA(pTs, Sc) + pCA_CA(pTs, Sc)) / 2;
}
double inline pTG_YA(double pTs, double Sc)
{
   return (pTG_TA(pTs, Sc) + pTG_CA(pTs, Sc)) / 2;
}
double inline pCA_YG(double pTs, double Sc)
{
   return (pCA_TG(pTs, Sc) + pCA_CG(pTs, Sc)) / 2;
}
double inline pCA_CR(double pTs, double Sc)
{
   return (pCA_CA(pTs, Sc) + pCA_CG(pTs, Sc)) / 2;
}
double inline pTG_YG(double pTs, double Sc)
{
   return (pTG_TG(pTs, Sc) + pTG_CG(pTs, Sc)) / 2;
}
double inline pCA_YR(double pTs, double Sc)
{
   return Sc * Sc / 4;
}
double inline pTG_YR(double pTs, double Sc)
{
   return Sc * Sc / 4;
}

/** ========================================================
 * Ref - reference sequence, which accumulates gaps from multiple
 * alignments of ungaped itself to other sequences.
 * seq - aligned portion of reference sequence.
 * offset - offset of the local alignment relative to the reference.
 * Offset must be 0-based.
 */
void adjustReference(string& ref, const string& seq, int offset)
{
   string::iterator sptr1 = ref.begin();
   string::const_iterator sptr2 = seq.begin();
   try
   {
      /**
       * Slide reference pointer to the beginning of the local alignment
       */
      for (; sptr1 != ref.end() && offset > 0; sptr1++)
      {
         if (isalpha(*sptr1) || *sptr1 == '*')
         {
            offset--;
         }
      }
      /**
       * Update reference with any new gaps from the alignment
       */
      while (sptr1 != ref.end() && sptr2 != seq.end())
      {
         if (*sptr1 != *sptr2)
         {
            if (*sptr1 == ScoreMatrix::GAPCHAR)
            {
               sptr1++;
               continue;
            }
            if (*sptr2 == ScoreMatrix::GAPCHAR)
            {
               sptr1 = ref.insert(sptr1, ScoreMatrix::GAPCHAR);
            }
         }
         sptr1++;
         sptr2++;
      }
   }
   catch (exception e)
   {
      cerr << e.what() << "\n";
   }
}
/** ========================================================
 offset - must be 0-based.
 ref0 - reference with accumulated gaps across multiple alignments;
 ref - aligned portion of reference from the particular paiwise alignment;
 seq - portion of a sequence aligned to the reference;
 offset - offset of the pairwise alignment relative to the reference;
 */
void adjustAlignment(const string& ref0, string& ref, string& seq, int offset)
{
   string::const_iterator sptr1 = ref0.begin();
   string::iterator sptr2 = ref.begin();
   string::iterator sptr3 = seq.begin();
   int insertLen = 0;
   try
   {
      for (; sptr1 != ref0.end() && offset > 0; sptr1++)
      {
         if (isalpha(*sptr1) || *sptr1 == '*')
         {
            offset--;
         }
         insertLen++;
      }
      while (sptr1 != ref0.end() && sptr2 != ref.end())
      {
         if (*sptr1 != *sptr2)
         {
            if (*sptr1 == ScoreMatrix::GAPCHAR)
            {
               sptr2 = ref.insert(sptr2, ScoreMatrix::GAPCHAR);
               sptr3 = seq.insert(sptr3, ScoreMatrix::GAPCHAR);
            }
            else
            {
               if (*sptr2 == ScoreMatrix::GAPCHAR)
               {
                  throw logic_error("ERROR: Unexpected gap symbol in MultipleAlignment::adjustAlignment");
               }
            }
         }
         sptr1++;
         sptr2++;
         sptr3++;
      }
      char filler = ScoreMatrix::LPADDING;
      ref.insert(0, insertLen, filler);
      seq.insert(0, insertLen, filler);
      filler = ScoreMatrix::RPADDING;
      ref.insert(ref.end(), ref0.end() - sptr1, filler);
      seq.insert(seq.end(), ref0.end() - sptr1, filler);
   }
   catch (exception e)
   {
      cerr << e.what() << "\n";
   }
}
